from __future__ import print_function

import matplotlib.pyplot as plt

import argparse
import os
import shutil
import time
import random
import numpy as np
import copy

os.environ['CUDA_VISIBLE_DEVICES'] = '0'

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.optim as optim
import torch.utils.data as data
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import models.cifar as models
from models.cifar.efficientnet import EfficientNet
from models.cifar.mobilenetv2 import mobilenetv2
from models.cifar.shufflenet import shufflenet
import torchvision.models as tv_models

from utils import Bar, Logger, AverageMeter, accuracy, mkdir_p, savefig

from tqdm import tqdm


model_names = sorted(name for name in models.__dict__
    if name.islower() and not name.startswith("__")
    and callable(models.__dict__[name]))

parser = argparse.ArgumentParser(description='PyTorch CIFAR10/100 Training')
# Datasets
parser.add_argument('-d', '--dataset', default='cifar10', type=str)
parser.add_argument('-j', '--workers', default=4, type=int, metavar='N',
                    help='number of data loading workers (default: 4)')
# Optimization options
parser.add_argument('--epochs', default=100, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('--train_batch', default=256, type=int, metavar='N',
                    help='train batchsize')
parser.add_argument('--test_batch', default=50, type=int, metavar='N',
                    help='test batchsize')
parser.add_argument('--lr', '--learning-rate', default=0.1, type=float,
                    metavar='LR', help='initial learning rate')
parser.add_argument('--drop', '--dropout', default=0, type=float,
                    metavar='Dropout', help='Dropout ratio')
parser.add_argument('--schedule', type=int, nargs='+', default=[150, 225],
                        help='Decrease learning rate at these epochs.')
parser.add_argument('--gamma', type=float, default=0.1, help='LR is multiplied by gamma on schedule.')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum')
parser.add_argument('--weight-decay', '--wd', default=5e-4, type=float,
                    metavar='W', help='weight decay (default: 1e-4)')
# Checkpoints
parser.add_argument('-c', '--checkpoint', default='checkpoint', type=str, metavar='PATH',
                    help='path to save checkpoint (default: checkpoint)')
parser.add_argument('--resume', default='', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
# Architecture
parser.add_argument('--arch', '-a', metavar='ARCH', default='vgg16_bn')
parser.add_argument('--depth', type=int, default=20, help='Model depth.')
parser.add_argument('--block-name', type=str, default='BasicBlock',
                    help='the building block for Resnet and Preresnet: BasicBlock, Bottleneck (default: Basicblock for cifar10/cifar100)')
parser.add_argument('--cardinality', type=int, default=8, help='Model cardinality (group).')
parser.add_argument('--widen-factor', type=int, default=4, help='Widen factor. 4 -> 64, 8 -> 128, ...')
parser.add_argument('--growthRate', type=int, default=12, help='Growth rate for DenseNet.')
parser.add_argument('--compressionRate', type=int, default=2, help='Compression Rate (theta) for DenseNet.')
# Miscs
parser.add_argument('--manualSeed', type=int, help='manual seed', default=1)
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                    help='evaluate model on validation set')
#Device options
parser.add_argument('--gpu-id', default='0', type=str,
                    help='id(s) for CUDA_VISIBLE_DEVICES')

parser.add_argument('--name', default='test')
parser.add_argument('--pruning_ratio', type=int, default=0)
parser.add_argument('--imp_iters', type=int, default=0)
parser.add_argument('--mode', type=str, default='lm', choices=['lm', 'sm', 'random', 'sign', 'magnitude'])
parser.add_argument('--reinit', action='store_true', default=False)
parser.add_argument('--learn_mask', action='store_true', default=False)
parser.add_argument('--lrr', action='store_true', default=False)
parser.add_argument('--diff_thr', type=float, default=0.9)
parser.add_argument('--warm_up', type=int, default=30)
parser.add_argument('--pretrained', type=str, default=None)
parser.add_argument('--gradual_prune', action='store_true', default=False)
parser.add_argument('--aws', action='store_true', default=False)

parser.add_argument('--lr_schedule', action='store_true', default=False)
parser.add_argument('--sign_transfer', action='store_true', default=False)
parser.add_argument('--mask_transfer', action='store_true', default=False)

parser.add_argument('--from_scratch', action='store_true', default=False)
parser.add_argument('--no_rewind', action='store_true', default=False)
parser.add_argument('--bn_init', action='store_true', default=False)
parser.add_argument('--no_fc_transfer', action='store_true', default=False)
parser.add_argument('--quantize', action='store_true', default=False)
parser.add_argument('--no_weight_rewind', action='store_true', default=False)

args = parser.parse_args()
state = {k: v for k, v in args._get_kwargs()}

# Validate dataset
#assert args.dataset == 'cifar10' or args.dataset == 'cifar100' or args.dataset == 'mnist'

# Use CUDA
os.environ['CUDA_VISIBLE_DEVICES'] = args.gpu_id
use_cuda = torch.cuda.is_available()

# Random seed
if args.manualSeed is None:
    args.manualSeed = random.randint(1, 10000)\

np.random.seed(args.manualSeed)
random.seed(args.manualSeed)
torch.manual_seed(args.manualSeed)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False
if use_cuda:
    torch.cuda.manual_seed_all(args.manualSeed)

best_acc = 0  # best test accuracy

def main():
    global best_acc
    start_epoch = args.start_epoch  # start from epoch 0 or last checkpoint epoch

    os.makedirs(f'checkpoint/{args.name}', exist_ok=True)

    # Data
    print('==> Preparing dataset %s' % args.dataset)
    if args.dataset.startswith('cifar'):
        transform_train = transforms.Compose([
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ])

        transform_test = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ])

        if args.dataset == 'cifar10':
            dataloader = datasets.CIFAR10
            num_classes = 10
        else:
            dataloader = datasets.CIFAR100
            num_classes = 100         
        
        args.train_batch = 128

        if args.arch =='resnet50':
            model = tv_models.resnet.resnet50(num_classes=num_classes)
            model.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
            model.maxpool = nn.Identity()
        if args.arch == 'mobilenetv2':
            model = mobilenetv2(num_classes=num_classes)
        if args.arch == 'shufflenet':
            model = shufflenet(num_classes=num_classes)
        if args.arch == 'efficientnet':
            from models.cifar.efficientnet import EfficientNetB0
            model = EfficientNetB0(num_classes)
        
        trainset = dataloader(root='./data', train=True, download=True, transform=transform_train)
        trainloader = data.DataLoader(trainset, batch_size=args.train_batch, shuffle=True, num_workers=8)

        testset = dataloader(root='./data', train=False, download=False, transform=transform_test)
        testloader = data.DataLoader(testset, batch_size=args.test_batch, shuffle=False, num_workers=8)

        final_epochs = 100

    elif args.dataset == 'tiny_imagenet':
        transform_train = transforms.Compose([
            transforms.RandomResizedCrop(64,(0.5,1)),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])

        transform_test = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])

        args.train_batch = 128

        num_classes = 200

        if args.arch =='resnet50':
            model = tv_models.resnet.resnet50(num_classes=num_classes)
            model.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
            model.maxpool = nn.Identity()
        if args.arch == 'mobilenetv2':
            model = mobilenetv2(num_classes=num_classes)
        if args.arch == 'shufflenet':
            model = shufflenet(num_classes=num_classes)
        if args.arch == 'efficientnet':
            from models.cifar.efficientnet import EfficientNetB0
            model = EfficientNetB0(num_classes)

        trainset = datasets.ImageFolder('tiny-imagenet-200/train', transform_train)

        trainloader = data.DataLoader(trainset, batch_size=args.train_batch, shuffle=True, num_workers=8)

        testset = datasets.ImageFolder('tiny-imagenet-200/val', transform_test)
        testloader = data.DataLoader(testset, batch_size=args.test_batch, shuffle=False, num_workers=8)

        final_epochs = 100
    
    elif args.dataset == 'imagenet':
        img_size = 224

        transform_train = transforms.Compose([
            transforms.RandomResizedCrop(img_size),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])

        transform_test = transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(img_size),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])


        num_classes = 1000

        if args.arch == 'mobilenetv2':
            model = tv_models.mobilenetv2.mobilenet_v2(num_classes=num_classes, dropout=0)
            args.train_batch = 512
        if args.arch == 'mlpmixer':
            from models.imagenet.mlp_mixer import MLPMixer
            model = MLPMixer(image_size=224, channels=3, patch_size=16, dim=512, depth=12, num_classes=1000)
            args.train_batch = 512

        trainset = datasets.ImageFolder('***yourdir***', transform_train)
        trainloader = data.DataLoader(trainset, batch_size=args.train_batch, shuffle=True, num_workers=16)

        testset = datasets.ImageFolder('***yourdir***', transform_test)
        testloader = data.DataLoader(testset, batch_size=args.test_batch, shuffle=False, num_workers=16)

        final_epochs = 120

    def new_conv_forward(self, x):
        if self.use_bias:
            return F.conv2d(x, self.weight * self.fixed_m.detach(), self.bias, stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups)
        else:
            return F.conv2d(x, self.weight * self.fixed_m.detach(), stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups)
    
    def new_conv1d_forward(self, x):
        if self.use_bias:
            return F.conv1d(x, self.weight, bias=self.bias, stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups) 
        else:
            return F.conv1d(x, self.weight * self.fixed_m.detach(), stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups)
    
    def new_linear_forward(self, x):
        if self.use_bias:
            return F.linear(x, self.weight * self.fixed_m.detach(), self.bias)
        else:
            return F.linear(x, self.weight * self.fixed_m.detach())
    
    def new_bn_forward(self, x):
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            assert self.num_batches_tracked is not None
            self.num_batches_tracked.add_(1)
            if self.momentum is None:  # use cumulative moving average
                exponential_average_factor = 1.0 / self.num_batches_tracked.item()
            else:  # use exponential moving average
                exponential_average_factor = self.momentum
        
        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is None)
        
        # Scale and shift
        gamma = (self.alpha + (1-self.alpha) * self.weight)
        beta = ((1-self.alpha) * self.bias)
        out = F.batch_norm(x, self.running_mean, self.running_var, gamma, beta, bn_training, exponential_average_factor, self.eps)
        
        return out
    def new_ln_forward(self, x):
        gamma = (self.alpha + (1-self.alpha) * self.weight)
        beta = ((1-self.alpha) * self.bias)
        out = F.layer_norm(x, self.weight.shape, gamma, beta)
        return out

    ii=0
    jj=0
    for m in model.modules():
        if hasattr(m, 'weight'):
            if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d) or isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.LayerNorm) or isinstance(m, nn.Conv1d):
                pass
            else:
                print(type(m))
                import pdb; pdb.set_trace()
        if isinstance(m, nn.Conv2d):
            if m.weight.shape[0] != num_classes and not hasattr(m, 'no_prune'):
                setattr(m, 'fixed_m', nn.Parameter(torch.ones(m.weight.shape),requires_grad=False))
                setattr(m, 'use_bias', m.bias is not None)
                setattr(m, 'sign_changes', nn.Parameter(torch.zeros(m.weight.shape).reshape(-1),requires_grad=False))
                new_method = new_conv_forward.__get__(m, m.__class__)
                setattr(m, 'forward', new_method)
            else:
                jj += 1
        if isinstance(m, nn.Linear) and not hasattr(m, 'no_prune'):
            if m.weight.shape[0] != num_classes:
                setattr(m, 'fixed_m', nn.Parameter(torch.ones(m.weight.shape),requires_grad=False))
                setattr(m, 'use_bias', m.bias is not None)
                setattr(m, 'sign_changes', nn.Parameter(torch.zeros(m.weight.shape).reshape(-1),requires_grad=False))
                new_method = new_linear_forward.__get__(m, m.__class__)
                setattr(m, 'forward', new_method)
            else:
                ii += 1

        if isinstance(m, nn.Conv1d) and not hasattr(m, 'no_prune'):
            setattr(m, 'fixed_m', nn.Parameter(torch.ones(m.weight.shape),requires_grad=False))
            setattr(m, 'use_bias', m.bias is not None)
            setattr(m, 'sign_changes', nn.Parameter(torch.zeros(m.weight.shape).reshape(-1),requires_grad=False))
            new_method = new_conv1d_forward.__get__(m, m.__class__)
            setattr(m, 'forward', new_method)

        if isinstance(m, nn.BatchNorm2d):
            setattr(m, 'alpha', 0)
            new_method = new_bn_forward.__get__(m, m.__class__)
            setattr(m, 'forward', new_method)
        if isinstance(m, nn.LayerNorm):
            setattr(m, 'alpha', 0)
            new_method = new_ln_forward.__get__(m, m.__class__)
            setattr(m, 'forward', new_method)
    
    assert ii==1 or jj==1

    model = model.cuda()
    cudnn.benchmark = True
    print('    Total params: %.2fM' % (sum(p.numel() for p in model.parameters())/1000000.0))
    criterion = nn.CrossEntropyLoss()

    initial_params = copy.deepcopy(model.state_dict())

    # Resume
    title = 'cifar-10-' + args.arch
    
    logger = Logger(os.path.join(f'checkpoint/{args.name}', 'log.txt'), title=title)
    logger.set_names(['Learning Rate', 'Train Loss', 'Valid Loss', 'Train Acc.', 'Valid Acc.'])

    if args.evaluate:
        print('\nEvaluation only')
        test_loss, test_acc = test(testloader, model, criterion, start_epoch, use_cuda)
        print(' Test Loss:  %.8f, Test Acc:  %.2f' % (test_loss, test_acc))
        return
    
    # Train and val
    for imp_iter in range(args.imp_iters+1):
        print(f'IMP ITERATION {imp_iter}/{args.imp_iters}')
        try: # load warmupped weights
            if imp_iter == 0 and 'warmup' not in args.name and args.pretrained is None and args.warm_up !=0:
                add = ''
                if args.aws:
                    add += '_aws'
                pretrained = torch.load(f'checkpoint/{args.dataset}_{args.arch}_warmup{args.warm_up}_seed{args.manualSeed}{add}/checkpoint_imp0_pr0.0_final.pth.tar')['state_dict']
                model.load_state_dict(pretrained)
                warmupped_params = copy.deepcopy(model.state_dict())
                continue
        except:
            print('no warmupped weights exist')
            import pdb; pdb.set_trace()
        
        if args.pretrained is not None:
            pretrained = torch.load(args.pretrained)['state_dict']
            model.load_state_dict(pretrained)

            if args.sign_transfer:
                sign_transfer(copy.deepcopy(model.state_dict()), initial_params, model, args.no_fc_transfer)
            elif args.mask_transfer:
                mask_transfer(copy.deepcopy(model.state_dict()), initial_params, model)

            total = 0
            pruned = 0
            for m in model.modules():
                if hasattr(m, 'fixed_m'):
                    total += m.weight.numel()
                    pruned += (m.fixed_m==0).float().sum()
            pr = pruned / total

            print('pr:', pr)

        epochs = args.epochs
        
        if args.dataset=='imagenet':
            optimizer = optim.Adam(model.parameters(), lr=0.001, weight_decay=1e-4)
            if args.lr_schedule:
                scheduler = optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=120, eta_min=0.00001)
            else:
                scheduler = optim.lr_scheduler.MultiStepLR(optimizer, milestones=[10000], gamma=0.1)
        else:
            milestones = [100000]
            if args.lr_schedule:
                milestones = [epochs//2,epochs//4*3]
            optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum, weight_decay=args.weight_decay)
            scheduler = optim.lr_scheduler.MultiStepLR(optimizer, milestones=milestones, gamma=0.1)

        for epoch in range(epochs):
            print(args.name)
            print('\nEpoch: [%d | %d]' % (epoch + 1, epochs))

            train_loss, train_acc = train(trainloader, model, criterion, optimizer, epoch, use_cuda, args)
            test_loss, test_acc = test(testloader, model, criterion, epoch, use_cuda)

            scheduler.step()

            print(f'Train Acc: {train_acc}\nTest Acc: {test_acc}')
            if args.aws and args.pretrained is None:
                print(f'During AWS, test accuracy can be substantially low due to the mismatch between running statistics and current batch statistics.')
                print(f'It will be recovered in the final training.')

            # append logger file
            logger.append([scheduler.get_lr()[0], train_loss, test_loss, train_acc, test_acc])

            # save model
            is_best = test_acc > best_acc
            best_acc = max(test_acc, best_acc)

            if args.pretrained is not None:
                logs = {'test acc': test_acc}

        logger.append([imp_iter,imp_iter,imp_iter,imp_iter,imp_iter])

        if imp_iter > 0 and args.pretrained is None:
            pr = prune(model, args.pruning_ratio, imp_iter, args.imp_iters, args.mode, args.learn_mask)
            #if args.sign_transfer:
            #    sign_transfer(copy.deepcopy(model.state_dict()), warmupped_params, model)
            #if args.mask_transfer:
            if not args.no_rewind:
                if args.no_weight_rewind:
                    mask_transfer(copy.deepcopy(model.state_dict()), initial_params, model)
                else:
                    mask_transfer(copy.deepcopy(model.state_dict()), warmupped_params, model)
        else:
            total = 0
            pruned = 0
            for m in model.modules():
                if hasattr(m, 'fixed_m'):
                    total += m.weight.numel()
                    pruned += (m.fixed_m==0).float().sum()
            pr = pruned / total
        
        print('pr', pr)

        save_checkpoint({
            'epoch': epoch + 1,
            'state_dict': model.state_dict(),
            'acc': test_acc,
            'best_acc': best_acc,
            'optimizer' : optimizer.state_dict(),
        }, False, checkpoint=f'checkpoint/{args.name}', filename=f'checkpoint_imp{imp_iter}_pr{pr}_final.pth.tar')

    logger.close()
    logger.plot()

def train(trainloader, model, criterion, optimizer, epoch, use_cuda, args):
    # switch to train mode
    model.train()

    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    top5 = AverageMeter()
    end = time.time()

    pbar = tqdm(trainloader)

    #bar = Bar('Processing', max=len(trainloader)//100)
    for batch_idx, (inputs, targets) in enumerate(pbar):
        # measure data loading time
        data_time.update(time.time() - end)

        if args.aws:
            alpha = random.random()
            for m in model.modules():
                if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.LayerNorm):
                    m.alpha = alpha
        else:
            for m in model.modules():
                if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.LayerNorm):
                    m.alpha = 0

        if use_cuda:
            inputs, targets = inputs.cuda(), targets.cuda()
        inputs, targets = torch.autograd.Variable(inputs), torch.autograd.Variable(targets)

        # compute output
        outputs = model(inputs)
        loss = criterion(outputs, targets)

        # measure accuracy and record loss
        prec1, prec5 = accuracy(outputs.data, targets.data, topk=(1, 5))
        losses.update(loss.item(), inputs.size(0))
        top1.update(prec1.item(), inputs.size(0))
        top5.update(prec5.item(), inputs.size(0))

        # compute gradient and do SGD step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        pbar.set_postfix(loss=losses.avg, acc=top1.avg)
    
    for m in model.modules():
        if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.LayerNorm):
            m.alpha = 0

    return (losses.avg, top1.avg)

def test(testloader, model, criterion, epoch, use_cuda):
    global best_acc

    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    top5 = AverageMeter()

    # switch to evaluate mode
    model.eval()

    end = time.time()

    feats = []
    labels = []
    for batch_idx, (inputs, targets) in enumerate(tqdm(testloader)):
        # measure data loading time
        data_time.update(time.time() - end)

        if use_cuda:
            inputs, targets = inputs.cuda(), targets.cuda()
        inputs, targets = torch.autograd.Variable(inputs, volatile=True), torch.autograd.Variable(targets)

        # compute output
        
        #outputs = model(inputs,True)
        #feats.append(outputs.detach().cpu())
        #labels.append(targets.cpu())
        #outputs = model.fc(outputs)

        outputs = model(inputs)
        loss = criterion(outputs, targets)

        # measure accuracy and record loss
        prec1, prec5 = accuracy(outputs.data, targets.data, topk=(1, 5))
        losses.update(loss.item(), inputs.size(0))
        top1.update(prec1.item(), inputs.size(0))
        top5.update(prec5.item(), inputs.size(0))

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

    #return torch.cat(feats,dim=0), torch.cat(labels,dim=0)

    return (losses.avg, top1.avg)

def save_checkpoint(state, is_best, checkpoint='checkpoint', filename='checkpoint.pth.tar'):
    filepath = os.path.join(checkpoint, filename)
    #os.makedirs(checkpoint, exist_ok=True)
    torch.save(state, filepath)
    if is_best:
        shutil.copyfile(filepath, os.path.join(checkpoint, 'model_best.pth.tar'))

def adjust_learning_rate(optimizer, epoch):
    global state
    if epoch in args.schedule:
        state['lr'] *= args.gamma
        for param_group in optimizer.param_groups:
            param_group['lr'] = state['lr']

def prune(model, pruning_ratio, imp_iter, total_iters, mode, learn_mask):
    total = 0
    pruned = 0
    alpha = pruning_ratio / 100
    weights = []
    for m in model.modules():
        if hasattr(m, 'fixed_m'):
            mask = m.fixed_m.reshape(-1)
            weights.append(m.weight.reshape(-1)[mask==1])
    
    weights = torch.cat(weights,dim=0)
    topk = abs(weights).topk(int(weights.shape[0]*(pruning_ratio/100)))[0][-1].cpu().detach().numpy()

    for m in model.modules():
        if hasattr(m, 'fixed_m'):
            weights = m.weight.data.cpu().reshape(-1).numpy()
            fixed_m = m.fixed_m.cpu().reshape(-1).numpy()

            num_total = weights.shape[0]
            fixed_m[fixed_m!=0] = (abs(weights)>=topk)[fixed_m!=0].astype(np.float32)

            new_mask = fixed_m.reshape(m.weight.shape)
        
            m.fixed_m.data = torch.from_numpy(new_mask).cuda()
            total+=num_total
            pruned+=(m.fixed_m==0).sum()

    print('pruned_ratio:',pruned / total)

    return pruned / total

def layerwise_prune(model, pruning_ratio, imp_iter, total_iters, mode, learn_mask):
    total = 0
    pruned = 0
    alpha = pruning_ratio / 100
    for m in model.modules():
        if hasattr(m, 'fixed_m'):
            weights = m.weight.data.cpu().reshape(-1).numpy()
            fixed_m = m.fixed_m.cpu().reshape(-1).numpy()

            num_total = weights.shape[0]
            num_to_prune = int(num_total*(1-alpha**imp_iter))
            sorted_indices = np.argsort(abs(weights*fixed_m))
            pruned_indices = sorted_indices[:num_to_prune]
            fixed_m[pruned_indices] = 0

            new_mask = fixed_m.reshape(m.weight.shape)
        
            m.fixed_m.data = torch.from_numpy(new_mask).cuda()
            total+=num_total
            pruned+=(m.fixed_m==0).sum()

    print('pruned_ratio:',pruned / total)

    return pruned / total
    
def initialize_randomness(seed):
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.cuda.manual_seed_all(seed)

def model_weight_reset(model):
    for m in model.modules():
        if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
            m.reset_parameters()

def sign_transfer(pretrained, init_sd, model, no_fc=False):
        model.load_state_dict(pretrained)

        signs = []
        b_signs = []
        masks = []

        bn_w = []
        bn_b = []

        c_w = []
        c_b = []

        for m in model.modules():
            if hasattr(m, 'fixed_m'):
                sign = torch.tensor(copy.deepcopy(torch.sign(m.weight*m.fixed_m).cpu().detach().numpy()))
                if m.use_bias:
                    b_sign = torch.tensor(copy.deepcopy(torch.sign(m.bias).cpu().detach().numpy()))
                    b_signs.append(b_sign)

                signs.append(sign)
            elif hasattr(m, 'no_prune'):
                c_w.append(copy.deepcopy(torch.sign(m.weight.data)))
                if m.bias is not None:
                    c_b.append(copy.deepcopy(torch.sign(m.bias.data)))
            elif isinstance(m, nn.Linear):
                fc_w_sign = torch.tensor(copy.deepcopy(torch.sign(m.weight).cpu().detach().numpy()))
                if m.bias is not None:
                    fc_b_sign = torch.tensor(copy.deepcopy(torch.sign(m.bias).cpu().detach().numpy()))

        model.load_state_dict(init_sd)

        i=0
        j=0
        ii_w=0
        ii_b=0
        iii = 0
        ic_w = 0
        ic_b = 0
        for m in model.modules():
            if hasattr(m, 'fixed_m'):
                m.weight.data = abs(m.weight.data) * signs[i].cuda()
                m.fixed_m.data = (signs[i].cuda() != 0)
                if m.use_bias:
                    m.bias.data = abs(m.bias.data) * b_signs[j].cuda()
                    j+=1
                i+=1
            elif hasattr(m, 'no_prune'):
                m.weight.data = abs(m.weight.data) * c_w[ic_w].cuda()
                ic_w += 1
                if m.bias is not None:
                    m.bias.data = abs(m.bias.data) * c_b[ic_b].cuda()
                    ic_b += 1
            elif isinstance(m, nn.Linear) and not no_fc:
                m.weight.data = abs(m.weight.data) * fc_w_sign.cuda()
                if m.bias is not None:
                    m.bias.data = abs(m.bias.data) * fc_b_sign.cuda()

def mask_transfer(pretrained, init_sd, model):
    model.load_state_dict(pretrained)

    masks = []
    for m in model.modules():
        if hasattr(m, 'fixed_m'):
            mask = torch.tensor(copy.deepcopy((m.fixed_m==1).cpu().detach().numpy()))
            masks.append(mask)

    model.load_state_dict(init_sd)

    i=0
    for m in model.modules():
        if hasattr(m, 'fixed_m'):
            m.fixed_m.data = masks[i].cuda()
            i+=1


if __name__ == '__main__':
    main()
